#ifndef MOC_VIDEO_QT_H
#define MOC_VIDEO_QT_H
 
#include <QWidget>
#include <QtWidgets>
#ifdef MPXPLAY_LINK_QTMEDIA
#include <QtMultimediaWidgets/QVideoWidget>
#endif
#include "moc_config.h"
#ifdef MPXPLAY_LINK_QTMEDIA
#include <QtMultimedia/QMediaPlayer>
#endif
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
//#include "evr/evrcustompresenter.h"
#ifdef __cplusplus
extern "C" {
#endif
#include <libavcodec/avcodec.h>
#include <libavutil/imgutils.h>
#include <libswscale/swscale.h>
#include <libavfilter/buffersink.h>
#include <libavfilter/buffersrc.h>
#include <libavutil/opt.h>
#ifdef __cplusplus
}
#endif // __cplusplus
#endif // MPXPLAY_LINK_ORIGINAL_FFMPEG

QT_BEGIN_NAMESPACE
class MainWindow;
class DispQtVideoEventFilter;
QT_END_NAMESPACE

//#define DISPQT_FFMPEG_TEST_SPEED 1 // test execution duration of different parts (rendering time)
//#define DISPQT_USE_OPENGL_WIDGET 1 // 1: doesn't work  2: slow
//#define DISPQT_USE_GDI_WIDGET 1    // slow, bad quality
//#define DISPQT_VIDEO_SCHEDULER_SEPARATED_AFFINITY 1 // set different affinity for (put on different cores) decoder and scheduler threads

#define DISPQT_VIDEO_OPEN_TIMEOUT  5000
#define DISPQT_VIDEO_MUTEX_TIMEOUT DISPQT_MUTEX_TIMEOUT  // 5000

#define DISPQT_VIDEO_CURSORHIDE_TIMEOUT 3000

#define DISPQT_VIDEOWIDGET_SINGLECLICK_TIMEOUTMS 300
#define DISPQT_VIDEOWIDGET_DOUBLECLICK_TIMEOUTMS 300

#define DISPQT_VIDEOWIDGET_CMENUHIDE_TIMEOUTMS 100

#define DISPQT_FFMPEG_WORKERTHREAD_REFRESH_MS 10

#define DISPQT_VIDEO_FILTERPROCESS_AVFORMAT AV_PIX_FMT_YUV420P

#define DISPQT_VIDEO_AUTOCROP_USE_RGB32 1  // currently the autocrop doesn't work properly in YUV

#if DISPQT_VIDEO_AUTOCROP_USE_RGB32
#define DISPQT_VIDEO_AUTOCROP_AVFORMAT AV_PIX_FMT_RGB32
#else
#define DISPQT_VIDEO_AUTOCROP_AVFORMAT DISPQT_VIDEO_FILTERPROCESS_AVFORMAT
#endif
#define DISPQT_VIDEO_PROCESS_BUFFERALIGNMENT 1   // with 16 : autocrop doesn't work properly (don't use 32, it seems to be buggy in FFmpeg -> segmentation fault)
#define DISPQT_VIDEO_QIMAGE_OUTPUT_AVFORMAT AV_PIX_FMT_RGB32
#define DISPQT_VIDEO_QIMAGE_OUTPUT_QFORMAT QImage::Format_RGB32
//#define DISPQT_VIDEO_QIMAGE_OUTPUT_AVFORMAT AV_PIX_FMT_BGRA      // there's no differene on the video surface
//#define DISPQT_VIDEO_QIMAGE_OUTPUT_QFORMAT QImage::Format_ARGB32 //

#define DISPQT_AVFRAME_FLAG_NOSKIPFRAME (1 << 30)  // AVFrame extra flag: disable (for visualization video frame)
#define DISPQT_AVFRAME_FLAG_AR_WINDOW   (1 << 31)  // AVFrame extra flag: use window aspect ratio (for visualization video frame)

#define DISPQT_VIDEOFILTER_INVALID_VALUE -1000000000  // used to identify the initial non variable_ptr ffmpeg elem config statuses

#define DISPQT_VIDEOFILTER_APPLYRETBIT_NOFILTERAPPLY       0
#define DISPQT_VIDEOFILTER_APPLYRETBIT_INPUTFRAME_PROCESSED (1 << 0)
#define DISPQT_VIDEOFILTER_APPLYRETBIT_OUTPUTFRAME_PRODUCED (1 << 1)

#define DISPQT_VIDEOFILTER_ELEMTYPE_BOOLEAN  (1 << 0)
#define DISPQT_VIDEOFILTER_ELEMTYPE_INTEGER  (1 << 1)
#define DISPQT_VIDEOFILTER_ELEMTYPE_FLOAT    (1 << 2)

#define DISPQT_VIDEOFILTER_PROCESSFLAG_DEFAULT   (1 << 0) // using of this filter is enabled if input/output formats differ and no other filter is enabled
#define DISPQT_VIDEOFILTER_PROCESSFLAG_DEPENDANT (1 << 1) // using of this filter is enabled only if other filters are used (pad at eq)
#define DISPQT_VIDEOFILTER_PROCESSFLAG_AUDIOSRC  (1 << 2) // audio source

#define DISPQT_VIDEOFILTER_NAMES_COLLECTION_BUFSIZE 512

//#define DISPQT_VIDEO_AUTOCROP_UPDATE          -1  // at crop mode change
#define DISPQT_VIDEO_AUTOCROP_COUNTER_REFRESH  1  // at seeking or at window resize
#define DISPQT_VIDEO_AUTOCROP_COUNTER_INIT     2  // at opening a new video

enum dispqt_videoframefilter_process_types
{
	DISPQT_VIDEOFRAMEFILTER_PROCESSTYPE_DECODER,
	DISPQT_VIDEOFRAMEFILTER_PROCESSTYPE_VMIXERSLIDER,
	DISPQT_VIDEOFRAMEFILTER_PROCESSTYPE_TRANSFORM,
	DISPQT_VIDEOFRAMEFILTER_PROCESSTYPE_MAX
};

enum dispqt_video_filter_crop_elems {
	DISPQT_VIDEO_FILTERCROP_ELEM_W = DISPQT_VIDEO_VMIXERTYPE_MAX,
	DISPQT_VIDEO_FILTERCROP_ELEM_H,
	DISPQT_VIDEO_FILTERCROP_ELEM_X,
	DISPQT_VIDEO_FILTERCROP_ELEM_Y,
	DISPQT_VIDEO_FILTERCROP_ELEM_MAX,
	DISPQT_VIDEO_FILTERCROP_ELEM_NUM = (DISPQT_VIDEO_FILTERCROP_ELEM_Y - DISPQT_VIDEO_FILTERCROP_ELEM_W + 1)
};

typedef struct dispqt_video_filter_format_s{
	int width;
	int height;
	enum AVPixelFormat pixelformat;
}dispqt_video_filter_format_s;

typedef struct dispqt_video_filter_info_s{
	AVFrame *ffmpeg_sws_frame;         // for autocrop and for rgb32 (qwidget) output if picture filtering is disabled
	AVFrame *ffmpeg_transformed_frame; // picture transformed (pad, rotation, etc.) frame
	AVFrame *ffmpeg_filtered_frame;    // picture filtered (brightness, contrast, etc.) frame
	mpxp_bool_t ffmpeg_swsframe_hasdata;      // ffmpeg_sws_frame has valid data
	struct SwsContext *sws_ctx;               // context of frame conversion
	struct dispqt_video_filter_format_s input_format_cfg;  // sws is initialized with this source format infos (came from the ffmpeg_decoded_frame)
	struct dispqt_video_filter_format_s output_format_cfg; // sws is initialized to this output format (sws output frame may has different pixelformat, like YUVJ to YUV change)
}dispqt_video_filter_info_s;

typedef struct dispqt_videofilter_elemdescriptor_s
{
	const char *ffmpeg_option_name;
	const int conf_value_offset;
	const int conf_value_scale;  // at DISPQT_VIDEOFILTER_ELEMTYPE_BOOLEAN, if *variable_ptr == conf_value_scale, then '1' is set, else '0'
	const int conf_value_center;
	int *variable_ptr;
	int last_configured_variable_value;
	const unsigned int flags;
	const int unique_id;
}dispqt_videofilter_elemdescriptor_s;

typedef struct dispqt_videofilter_processdescriptor_s
{
	const char *ffmpeg_filter_name;
	const unsigned int process_type;
	const unsigned int process_flags;
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	unsigned int (*process_condition_func)(struct mmc_dispqt_config_s *gcfg, AVFrame *inframe, char *filtername_withconfig, unsigned int fnwcbuf_size);
#else
	unsigned int (*process_condition_func)(struct mmc_dispqt_config_s *gcfg);
#endif
	struct dispqt_videofilter_elemdescriptor_s *elemdesc;
}dispqt_videofilter_processdescriptor_s;

class DispQtVideoSurfaceContextMenu : public DispQtDialogElemQmenu
{
Q_OBJECT
public:
	DispQtVideoSurfaceContextMenu(MainWindow *main_window = 0, DispQtVideoEventFilter *parent = 0, QWidget *filtered = 0);
	~DispQtVideoSurfaceContextMenu();
protected:
	void mouseReleaseEvent(QMouseEvent *mevent);

private Q_SLOTS:
	void play_start_or_pause(void);
	void program_exit(void);
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	void actfunc_audio_visualization_select(QAction *selected_visualization_action);
	void actfunc_file_add_filestreams_to_video(void);
	void actfunc_file_add_livestream_to_video(void);
	void actfunc_file_record_streams_start(void);
	void actfunc_file_record_streams_stop(void);
	void actfunc_video_program_select(QAction *selected_ar_action);
	void actfunc_video_stream_select(QAction *selected_ar_action);
	void actfunc_video_ar_select(QAction *selected_ar_action);
	void actfunc_video_crop_select(QAction *selected_crop_action);
	void actfunc_video_deinterlace_select(QAction *selected_deinterlace_action);
	void actfunc_video_rotation_select(QAction *selected_rotation_action);
	void actfunc_video_zoom_select(QAction *selected_zoom_action);
	void actfunc_video_videowall_switch(void);
#endif

private:
	MainWindow *main_window;
	DispQtVideoEventFilter *parent_object;
	QWidget *filtered_widget;
	struct mpxplay_programinfo_t *program_list;
	struct mpxplay_streaminfo_t *stream_list;

	DispQtDialogElemQmenu *videocontmenu_sub_audio;
	DispQtDialogElemQmenu *videocontmenu_subaudio_sub_visualization;
	DispQtDialogElemQmenu *videocontmenu_sub_file;
	DispQtDialogElemQmenu *videocontmenu_sub_programs;
	DispQtDialogElemQmenu *videocontmenu_sub_streams;
	DispQtDialogElemQmenu *videocontmenu_sub_video;
	DispQtDialogElemQmenu *videocontmenu_subvideo_sub_aspectratio;
	DispQtDialogElemQmenu *videocontmenu_subvideo_sub_crop;
	DispQtDialogElemQmenu *videocontmenu_subvideo_sub_deinterlace;
	DispQtDialogElemQmenu *videocontmenu_subvideo_sub_rotation;
	DispQtDialogElemQmenu *videocontmenu_subvideo_sub_zoom;

	QAction *menuaction_filename, *action_play_start, *action_play_stop;
	QAction *action_play_skipback, *action_play_skipforward, *action_view_videofs, *action_show_epgdialog, *action_file_exit;
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	QAction *actions_audio_visualization_select[DISPQT_AUDIOVISUALIZATIONTYPE_MAX];
	QAction *action_file_addfilestreamstovideo, *action_file_addlivestreamstovideo, *action_file_recordstreams;
	QAction **actions_program_select;
	QAction **actions_stream_select[MPXPLAY_STREAMTYPEINDEX_NUM];
	QAction *actions_video_ar_select[DISPQT_VIDEOFULLSCREENARTYPE_MAX];
	QAction *actions_video_crop_select[DISPQT_VIDEOSCRCROPTYPE_MAX];
	QAction *actions_video_deinterlace_select[DISPQT_VIDEO_DEINTERLACETYPE_MAX];
	QAction *actions_video_rotation_select[DISPQT_VIDEO_ROTATIONTYPE_MAX];
	QAction *actions_video_zoom_select[DISPQT_VIDEOFULLSCREENZOOMTYPE_MAX];
	QAction *action_video_videowall;
#endif
};

class DispQtVideoEventEPGTooltip : public DispQtDialogElemWindow
{
Q_OBJECT
public:
	DispQtVideoEventEPGTooltip(MainWindow *mainwindow = 0, QWidget *parent = 0, QObject *eventfilter = 0, const QString &EPGeventList = 0, bool isVideoWall = false);

protected:
	bool event(QEvent *event);
	void mouseDoubleClickEvent(QMouseEvent *event);
	void keyPressEvent(QKeyEvent *event);

private:
	MainWindow *main_window;
	QWidget *parent_filtered_widget;
	QObject *parent_event_filter;
	bool is_videowall;
};

class DispQtVideoEventFilter : public QObject
{
Q_OBJECT
public:
	DispQtVideoEventFilter(MainWindow *main_window = 0, QWidget *parent = 0, QWidget *filtered = 0);
	bool get_mousecursor_forbidden_hide(void) { return this->mouse_cursor_forbidden_hide; }
	void mouse_cursor_show_timed(bool timed);
	void mouse_cursor_forbid_hide(bool forbid);
	void video_surface_move(int move_x, int move_y);
	void video_zoom_step(int globalpos_x, int globalpos_y, int direction); // 1: zoomin, -1:zoomout
	void video_callback_call(int globalpos_x, int globalpos_y, unsigned long command, mpxp_ptrsize_t arg1, mpxp_ptrsize_t arg2);
	DispQtVideoSurfaceContextMenu *video_surface_contextmenu;

private Q_SLOTS:
	void mouse_cursor_hide(void);
	void cmenu_about_hide(void);
	void mouse_singleclick_timeouthandler(void);
	void mouse_epgtooltip_presenter_handler(void);
	void mouse_epgtooltip_close(void);

protected:
	bool eventFilter(QObject *watched, QEvent *event);

private:
	bool is_video_fullscreen(void);
	bool is_video_wall_enabled(void);
	void videosurface_get_pos(QMouseEvent *me, int *pos_x, int *pos_y);
	void videosurface_handle_singleclick(void);
	void videosurface_handle_doubleclick(void);
	void videosurface_handle_moveevent_nobutton(QMouseEvent *me);
	void mouse_epgtooltip_open(int pos_x, int pos_y, int height, char *epg_lst_collected_str, bool is_videwall);
	MainWindow *main_window;
	QWidget *parent_widget, *filtered_widget;
	QString epg_event_list;
	DispQtVideoEventEPGTooltip *epgtooltip_window;
	Qt::MouseButton mouse_pressed_buttons;
	int mouse_last_globalpos_x, mouse_last_globalpos_y;
	int epgtooltip_globalpos_x, epgtooltip_globalpos_y;
	bool mouse_cursor_forbidden_hide, singleclick_event, singleclick_fullsrc_start, fullsrc_curr, is_cursor_overrided;
	QTimer timer_singleclick, timer_doubleclick, timer_mouse_hide, timer_cmenu_hide, timer_epgtooltip_interval, timer_epgtooltip_show;
	QMutex mutex_video;
};

class DispQtVideoFrameFilter : public QObject
{
Q_OBJECT
public:
	DispQtVideoFrameFilter(MainWindow *main_window = 0, QObject *parent = 0, dispqt_videofilter_processdescriptor_s *process_desc = 0, unsigned int nb_procdesc = 0);
	~DispQtVideoFrameFilter();

	int video_filter_close(void);
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	int video_filter_check(AVFrame *frame_in, int frameout_format);
	int video_filter_apply(AVFrame *frame_in, AVFrame **frame_out, int frameout_format);
#endif
	int getVideoFilterValue(int index);

public Q_SLOTS:
	void resetVideoFilterValues(int process_type);
	void setVideoFilterValue(int index, int value);

private:
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	void video_filter_opencl_init(void);
	int video_filter_init(int width, int height, int pixeltype_in, int pixeltype_out, AVFrame *inframe);
	int video_filter_config(int width, int height, int pixeltype_in, int pixeltype_out, AVFrame *inframe);
#endif
	dispqt_videofilter_elemdescriptor_s *searchElemByUniqueId(int id, int *process_type);
	bool setVideoFilterInternalValue(int index, int value);

	MainWindow *main_window;
	QObject *parent_object;
	QMutex mutex_filter;
	dispqt_videofilter_processdescriptor_s *videofilter_process_descriptor;
	unsigned int nb_video_filter_process_descriptors;
	dispqt_videofilter_processdescriptor_s *first_selected_process_descriptor;
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	int videofilter_initialized_width;
	int videofilter_initialized_height;
	int videofilter_initialized_pixeltype;
	AVFilterContext *videofilter_ffmpeg_buffersrc_ctx;
	AVFilterContext *videofilter_ffmpeg_buffersink_ctx;
	AVFilterGraph *videofilter_ffmpeg_graph;
	AVBufferRef *opencl_device_ref;
	char videofilter_ffmpeg_filternames_collection[DISPQT_VIDEOFILTER_NAMES_COLLECTION_BUFSIZE];
#endif
};

#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG

#define INFFMPEG_ASS_DECODED_MAX_LINES    8
#define INFFMPEG_ASS_DECODED_FLAG_ITALIC (1 << 0)

#define MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_DEFAULT    100
#define MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT      10
#define MPXPLAY_VIDEO_WORKER_THREAD_CRITICAL_ERROR_COUNT 10

#define INFFMPEG_VIDEO_RENDER_MIN_INTERVAL_MS  (DISPQT_FFMPEG_WORKERTHREAD_REFRESH_MS / 2)   // we never display more than 200 fps
#define INFFMPEG_VIDEO_RENDER_MAX_INTERVAL_MS  1000  // we try to display min. 1 fps in any case (other frames are dropped)
#define INFFMPEG_VIDEO_INITERROR_LIMIT 100  // ca. 1000ms wait with 10ms sleep

#define MPXPLAY_VIDEO_WORKDEC_FILTER_RESTORE_DELAY 250 // for video_delay_counter_workdec_filter (in frames) (10 sec at 25 fps)

typedef struct ffmpegvideo_subtitle_info_s
{
	int last_render_x, last_render_y, last_size_x, last_size_y;
	unsigned int nb_lines;
	char *line_texts[INFFMPEG_ASS_DECODED_MAX_LINES];
	unsigned int line_flags[INFFMPEG_ASS_DECODED_MAX_LINES];
}ffmpegvideo_subtitle_info_s;

typedef struct ffmpegvideo_decoder_s{
	unsigned int flags;                                          // DISPQT_FFMPGVIDEOCALLBACKFLAG_
	unsigned int streamtype_index;                               // MPXPLAY_STREAMTYPEINDEX_
	mpxplay_packetlist_t *video_pktlist_elem;                    // input video frame
	AVFrame *vid_frame_decoded;				                     // decoded video frame in video_workdec.cpp
	int video_last_valid_frame_w, video_last_valid_frame_h;      // for progressive jpeg (to drop small frames)
	int videoctx_init_error_count;
	int video_delay_counter_workdec_filter;                      // wait frames after filter is switched off
	int video_decoder_lost_frames;                               // number of pushed, but not decoded video frames (to switch off KeyframesOnly at videowall)
	mpxp_int64_t video_dec_last_timestamp;
	mpxplay_packetqueue_t videoframe_queue;                      // output frame queue
	mpxplay_packetlist_t *videoframe_pktlist_elem;               // output frame packet
	AVCodecContext *video_ctx;                                   // opened video context
	AVFrame *video_frame_decoded_preprocessed;                   // produced output video frame by video_workdec (deinterlaced, tonemapped) (old vid_frame_filtered)
	bool resetvideo;                                             // seek
	bool resetfilter;                                            //
	bool resetoutqueue;                                          //
	bool resetcodec;                                             // video codec (possible resolution) change
	DispQtVideoFrameFilter *video_frame_filter_workerprocess;    // processing filters (deinterlace)
	int videowall_window_pos_x, videowall_window_pos_y;          // video window x,y position on the videowall
	int videowall_window_width, videowall_window_height;         // video window size on the videowall
	dispqt_video_filter_info_s frame_conv_infos;                 // resize or change pixelformat (seekpreview)
}ffmpegvideo_decoder_s;

typedef struct ffmpegvideo_callback_command_s{
	unsigned int cmd;
	mpxp_ptrsize_t arg1, arg2;
}ffmpegvideo_callback_command_s;

typedef struct ffmpegvideo_callback_list_s{
	unsigned int flags;                                          // DISPQT_FFMPGVIDEOCALLBACKFLAG_
	mpxplay_infile_callback_t *infile_callback;
	mpxp_ptrsize_t infile_passdata;
	int video_init_framecounter;                                 // aka 'firstframes' flag, process video delay
	int video_delay_type;                                        // DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_
	int video_delay_large_counter;                               // to control video frame drop (decoder + output) with delay
	int video_delay_framedrop_counter;                           // belongs to the decoded output frames
	mpxp_int64_t video_last_timestamp;
	mpxp_int64_t video_last_render_time;
//	mpxp_int64_t video_last_render_duration;                     // unused yet (to detect slow rendering)
	mpxp_int64_t subtitle_begin_timestamp;                       // decoded subtitle display time begin (match with the video timestamp)
	mpxp_int64_t subtitle_end_timestamp;                         // decoded subtitle time end
	mpxplay_packetlist_t *subtitle_pktlist_elem;                 // subtitle in data
	ffmpegvideo_subtitle_info_s *subtitle_infos;                 // subtitle out/decoded data
	mpxp_thread_id_type videodecoder_thread_id;                  // decoder thread id
	DispQtVideoFrameFilter *video_frame_filter_workerprocess;    // processing filters (deinterlace)
	mpxp_int32_t selected_stream_index;                          // main video stream selection by streamtype_index from the video_decoder_threads
	bool is_videowall_enabled;                                   // use mosaic playing of parallel video streams
	bool is_videowall_disabled_by_decoder;                       // videowall is disabled by decoder (playing audio/radio program without video content)
	int video_wall_artype;                                       // aspect ratio mode of videowall output screen
	int videowall_nb_streams;                                    // number of videos (programs) displayed on the videowall (only the visible streams)
	int videowall_count_streams;                                 // counter for videowall streams (to update nb_streams inline)
	int videowall_nb_windows_x;                                  // number of videos (programs) displayed on the videowall horizontally
	int videowall_nb_windows_y;                                  // number of videos (programs) displayed on the videowall vertically
	int videowall_output_artype;                                 // aspect ratio mode for videowall (derived from video_wall_artype)
	ffmpegvideo_decoder_s video_decoder_threads[MPXPLAY_STREAMTYPEINDEX_MAX];
	ffmpegvideo_callback_command_s callback_command;
	struct ffmpegvideo_callback_list_s *next;
	mpxp_int64_t videowall_sync_timestamp;                       // sync timestamp of primary video to control/gain other videowall streams
	dispqt_video_filter_format_s videowall_subwindow_format;     // output size and format of a videowall window (sub-picture)
	dispqt_video_filter_format_s videowall_output_format;        // output size and format of the complete videowall
#ifdef DISPQT_FFMPEG_TEST_SPEED
	mpxp_int64_t timedevent_prev_sync_time;
#endif
}ffmpegvideo_callback_list_s;

extern void mpxplay_video_ffmpeg_subtitleinfo_clear(ffmpegvideo_subtitle_info_s *sub_lines);
extern void mpxplay_video_ffmpeg_subtitleinfo_copy(ffmpegvideo_subtitle_info_s *dest_lines, ffmpegvideo_subtitle_info_s *sub_lines);
extern ffmpegvideo_subtitle_info_s *mpxplay_decoderworker_subtitle_get(ffmpegvideo_callback_list_s *cbelem, mpxp_int64_t timestamp_curr_us);
extern void mpxplay_decoderworker_subtitle_reset(ffmpegvideo_callback_list_s *cbelem);

#define DISPQT_FFMPGVIDEO_MAX_QUEUED_VIDEOFRAME  5
#define DISPQT_FFMPGVIDEO_MAX_QUEUED_VISVIDEOFRAME  (DISPQT_FFMPGVIDEO_MAX_QUEUED_VIDEOFRAME * 5)
#define DISPQT_FFMPGVIDEO_MAX_DELAY_VISVIDEOFRAME   (DISPQT_FFMPGVIDEO_MAX_QUEUED_VIDEOFRAME * 4) // max number of video frames difference between audio and video output (else ignore timestamp (skip visual video frame))
#define DISPQT_FFMPGVIDEO_VIDEO_MAX_FRAME_DIFF_US   (15 * AV_TIME_BASE)  // max 15 secs difference between video frames
#define DISPQT_FFMPGVIDEO_AVSYNC_MAX_VALID_DIFF_US  (60 * AV_TIME_BASE) // probably an incorrect pts received, display output picture (ignore timestamp)
#define DISPQT_FFMPGVIDEO_SUBTITLE_MAX_VIDEO_DIFF_US (10 * AV_TIME_BASE)  // max 10 secs difference between video and subtitle (else don't display subtitle)

#define DISPQT_FFMPGVIDEOCALLBACK_INITFRAMECOUNTER_START 50 // video frames

#define DISPQT_FFMPGVIDEOCALLBACKFLAG_EXITEVENT           (1 << 0)  // stop thread event
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_THEXITDONE          (1 << 1)  // done
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_FIRSTFRAME          (1 << 4)  // first displayed frame sign
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO        (1 << 5)  // primary video file
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_DISPLAYONEFRAME     (1 << 6)  // display one video frame (in pause)
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP     (1 << 7)  // ignore timestamp sync (after seeking) to display a frame
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_STILLPICT           (1 << 8)  // still picture (aspect ratio and multiply picture frame workaround)
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_REFRESHTYPE_UPDATE  (1 << 9)  // no new output video frame, just refresh the output (reqest for DISPQT_VIDEOSURFACE_REFRESHTYPE_UPDATE at the case of resize/zomm/move/crop window changes)
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_AVAILABLE        (1 << 10) // video has multiply programs, videowall is available
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_ONLY   (1 << 11) // decode only keyframes at non-primary videowall windows
#define DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_NO     (1 << 12) // disable decode only keyframes for this window (keyframe decoding run to timeout by video_decoder_lost_frames)

enum {DISPQT_VIDEOSCHEDULER_CONTROL_RUN = 0, DISPQT_VIDEOSCHEDULER_CONTROL_REQ_SUSPEND, DISPQT_VIDEOSCHEDULER_CONTROL_REQ_CLOSE, DISPQT_VIDEOSCHEDULER_STATUS_SUSPENDED, DISPQT_VIDEOSCHEDULER_STATUS_CLOSED};
enum {DISPQT_VIDEOSURFACE_REFRESHTYPE_NONE = 0, DISPQT_VIDEOSURFACE_REFRESHTYPE_UPDATE, DISPQT_VIDEOSURFACE_REFRESHTYPE_RESET, DISPQT_VIDEOSURFACE_REFRESHTYPE_CLEAR, DISPQT_VIDEOSURFACE_REFRESHTYPE_CLOSE};

enum {DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_NONE = 0, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SMALL, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_MIDDLE, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_LARGE, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SIGNIFICANT};
enum {DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYLIMIT_SMALL = 250000, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYLIMIT_MIDDLE = 500000, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYLIMIT_LARGE = 1000000, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYLIMIT_SIGNIFICANT = 3000000};
enum {DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYCOUNT_MIDDLE = 3, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYCOUNT_LARGE = 10, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYCOUNT_SIGNIFICANT = 20};
enum {DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYFRAMEDROP_NONE = 0, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYFRAMEDROP_LARGE = 1, DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYFRAMEDROP_SIGNIFICANT = 1};

typedef struct mpxplay_dispqt_ffmpegvideo_worker_info_s{
	ffmpegvideo_callback_list_s *cblist_first, *cblist_last;
	int nb_callbacks;
#ifdef DISPQT_FFMPEG_TEST_SPEED
	mpxp_int64_t timedevent_prev_start_time;
#endif
	void *mutexhnd_workerinfo;
	bool is_video_stream, is_videosurface_open, is_paused;
	bool is_videowall_enabled;
	unsigned int video_scheduler_control, video_scheduler_status;
	unsigned int video_surface_refresh_request;
	MainWindow *main_window;
}mpxplay_dispqt_ffmpegvideo_worker_info_s;

class FFMpegVideoDecoderWorker : public QObject
{
Q_OBJECT
public:
	FFMpegVideoDecoderWorker(MainWindow *main_window = 0);
	~FFMpegVideoDecoderWorker();
	bool videothread_ffmpeg_config_callback(unsigned int cfgcmd, mpxp_ptrsize_t cbk_func, mpxp_ptrsize_t passdata);
	void videothread_ffmpeg_callback_call(int globalpos_x, int globalpos_y, unsigned long command, mpxp_ptrsize_t arg1, mpxp_ptrsize_t arg2);
	bool decoderworker_video_scheduler_control(unsigned int control_req);
	void decoderworker_video_surface_refresh(unsigned int refresh_type); // DISPQT_VIDEOSURFACE_REFRESHTYPE_

private:
	void cb_queue_videostreamdecoderthread_set(ffmpegvideo_callback_list_s *cbelem, int streamtype_index);
	void cb_queue_videostreamdecoderthread_clear(ffmpegvideo_callback_list_s *cbelem, int streamtype_index);
	int  cb_queue_put_entry(struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos, mpxp_ptrsize_t cbk_func, mpxp_ptrsize_t passdata, bool insert_to_begin);
	int  cb_queue_del_entry(struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos, mpxp_ptrsize_t cbk_func, mpxp_ptrsize_t passdata);
	void cb_queue_elem_clear(struct ffmpegvideo_callback_list_s *cbelem);
	void cb_queue_clear(struct mpxplay_dispqt_ffmpegvideo_worker_info_s *workinfos);

	int  videothread_videowall_search_windowitem(ffmpegvideo_callback_list_s *cbelem, int streamtypeindex, int globalpos_x, int globalpos_y, bool halve_win_width);
	int  decoderworker_video_load_next_frame(ffmpegvideo_callback_list_s *cbelem, bool *resetvideo);
	void decoderworker_video_reset(ffmpegvideo_callback_list_s *cbelem, int video_index);
	bool decoderworker_video_frame_scheduler(ffmpegvideo_callback_list_s *cbelem, int video_index, mpxp_int64_t timestamp_curr_ms);

	MainWindow *main_window;
	mpxplay_dispqt_ffmpegvideo_worker_info_s video_worker_infos;
	mpxp_thread_id_type videodworker_thread_id;
};

//-------------------------------------------------------------------------------------------------------------

typedef struct dispqt_video_source_info_s{
	AVFrame *ffmpeg_decoded_frame;     // came from the decoder (decoded, sw deinterlaced frame), in DISPQT_VIDEO_FILTERPROCESS_AVFORMAT, or in native decoder output
	AVRational video_aspect_ratio;     // to detect aspect ratio changes (recalc surface)
	int sws_srccrop_x, sws_srccrop_y, sws_srccrop_width, sws_srccrop_height; // crop of the source (manual or auto)
	unsigned long last_video_crop_type, ffmpeg_swsframe_cropreset;
	int sws_autoxcrop_framecounter;
	mpxp_bool_t update_crop;
	bool check_autocrop;
}dispqt_video_source_info_s;

typedef struct dispqt_video_surface_info_s{
	int sws_crop_x, sws_crop_y, sws_crop_w, sws_crop_h;                      // crop of the output (zoom / move)
	int sws_dst_width, sws_dst_height, sws_dst_pos_x, sws_dst_pos_y;
	int window_pos_x, window_pos_y, window_size_x, window_size_y;
	int last_video_zoom_min_val, last_video_zoom_ti_val;
	int video_zoom_globalpos_x, video_zoom_globalpos_y;
	//unsigned long last_video_render_type;
#if (DISPQT_USE_OPENGL_WIDGET == 1)
	QOpenGLContext *video_opengl_context;
	QSurfaceFormat *video_opengl_sfmt;
	QOpenGLPaintDevice *video_opengl_device;
	QWindow *topwindow_hnd;
	//QOffscreenSurface *video_opengl_surface;
	mpxp_bool_t video_opengl_initialized;
#endif
//	Scheduler *video_sample_scheduler_evr;
//	EVRCustomPresenter *video_renderer_evr;
	AVFrame *ffmpeg_output_frame;                // points to ffmpeg_sws_frame or ffmpeg_filtered_frame after preprocess
	int video_outframe_last_width, video_outframe_last_height;  // last displayed video frame width / height (to detect resolution changes -> recalc surface)
	mpxp_bool_t videoout_surface_recalc;         // re-calculate surface values (videowidget_ffmpeg_swsctx_calculate)
	mpxp_bool_t videoout_surface_clearpic;       // clear output surface before or instead of rendering (eg: at autocrop, or at new file)
	enum AVPixelFormat videoout_surface_pix_fmt; // the final output format (RGB32 for qwidget)
	mpxp_bool_t update_resize, do_resize_repaint;
	int updateresize_x, updateresize_y;
//	mpxp_bool_t update_move;
//	int updatemove_x, updatemove_y;
//	mpxp_bool_t update_zoom;
//	int updatezoom_center_x, updatezoom_center_y, updatezoom_direction;
	mpxp_bool_t update_filter_trans, update_filter_eq; // basically they belong to the filter
}dispqt_video_surface_info_s;

extern int  mpxplay_dispqt_videotrans_ffmpeg_swsctx_init(struct dispqt_video_filter_info_s *video_filter_infos, AVFrame *input_video_frame, struct dispqt_video_filter_format_s *output_format);
extern void mpxplay_dispqt_videotrans_ffmpeg_swsctx_clean(struct dispqt_video_filter_info_s *video_filter_infos);
extern void mpxplay_dispqt_videotrans_ffmpeg_swsctx_close(struct dispqt_video_filter_info_s *video_filter_infos);
extern bool mpxplay_dispqt_videotrans_ffmpeg_swsctx_conv_frame(struct dispqt_video_filter_info_s *video_filter_infos, AVFrame *ivf, struct dispqt_video_filter_format_s *output_format);

extern int mpxplay_dispqt_videowidget_ffmpeg_swsctx_get_custom_ar(char *custom_str, int scale_value, int default_value);

class FFMpegVideoSoftwareTransform
{
public:
	FFMpegVideoSoftwareTransform(MainWindow *main_window = 0, dispqt_video_source_info_s *video_source_infos = 0, dispqt_video_surface_info_s *video_surface_infos = 0);
	~FFMpegVideoSoftwareTransform();

	void videowidget_ffmpeg_videoctx_clear(void);
	void videowidget_ffmpeg_swsctx_close(void);

	void videowidget_ffmpeg_swsctx_crop_auto_reset(void);
	void videowidget_ffmpeg_swsctx_crop_auto_clear(void);

	void videowidget_ffmpeg_filtered_crop_update(bool check_autocrop);
	void videowidget_ffmpeg_filtered_trans_update(void);
	void videowidget_ffmpeg_filtered_eq_update(void);

private:
	int  videowidget_ffmpeg_videoctx_init(struct dispqt_video_filter_info_s *video_filter_infos);
	void videowidget_ffmpeg_videoctx_close(struct dispqt_video_filter_info_s *video_filter_infos);
	bool videowidget_ffmpeg_swsctx_crop_auto_detect(AVFrame *input_video_frame, dispqt_video_source_info_s *source_infos, mpxp_bool_t *videooutsurface_clearpic);
	int  videowidget_ffmpeg_swsctx_init(struct dispqt_video_filter_info_s *video_filter_infos, AVFrame *input_video_frame, int output_width, int output_height, enum AVPixelFormat output_pixelformat);
	int  videowidget_ffmpeg_swsctx_get_sws_outformat(struct dispqt_video_filter_info_s *video_filter_infos, AVFrame *input_video_frame);
	bool videowidget_ffmpeg_swsctx_conv_frame(struct dispqt_video_filter_info_s *video_filter_infos);
	bool videowidget_ffmpeg_swsctx_check_swsframe_for_outsurface(struct dispqt_video_filter_info_s *video_filter_infos);
	void videowidget_ffmpeg_swsctx_close(struct dispqt_video_filter_info_s *video_filter_infos);
	void videowidget_ffmpeg_swsctx_clear(struct dispqt_video_filter_info_s *video_filter_infos);

	MainWindow *main_window;
	dispqt_video_source_info_s *video_source_infos;
	dispqt_video_surface_info_s *video_surface_infos;
	dispqt_video_filter_info_s video_filter_infos;
};

#if (DISPQT_USE_OPENGL_WIDGET == 2)
class FFMpegVideoWidget : public QOpenGLWidget
#else
class FFMpegVideoWidget : public QWidget
#endif
{
Q_OBJECT
public:
	FFMpegVideoWidget(MainWindow *main_window = 0, QWidget *parent = 0);
	~FFMpegVideoWidget();
//	bool event(QEvent *event);
	bool videowidget_ffmpeg_config_callback(unsigned int cfgcmd, mpxp_ptrsize_t cbk_func, mpxp_ptrsize_t passdata);
	void videowidget_ffmpeg_callback_call(int globalpos_x, int globalpos_y, unsigned long command, mpxp_ptrsize_t arg1, mpxp_ptrsize_t arg2);
	bool videowidget_ffmpeg_schedulerthread_suspend(void);
	void videowidget_ffmpeg_schedulerthread_resume(void);

Q_SIGNALS:
	void signal_ffmpeg_swctx_resize(int window_width, int window_height, bool do_repaint);
	void signal_ffmpeg_swctx_move(int move_x, int move_y);
	void signal_ffmpeg_swctx_zoom(int globalpos_x, int globalpos_y, int direction);
	void signal_ffmpeg_filtered_trans_refresh(void);
	void signal_ffmpeg_filtered_eq_refresh(void);

public Q_SLOTS:
	void videowidget_ffmpeg_swctx_resize(int window_width, int window_height, bool do_repaint);
	void videowidget_ffmpeg_swctx_move(int move_x, int move_y);
	void videowidget_ffmpeg_swctx_zoom(int globalpos_x, int globalpos_y, int direction);
	void videowidget_ffmpeg_filtered_crop_refresh(bool check_autocrop);
	void videowidget_ffmpeg_filtered_trans_refresh(void);
	void videowidget_ffmpeg_filtered_eq_refresh(void);
	void videowidget_ffmpeg_frame_display(int video_index, unsigned int flags, AVFrame *videoout_frame, ffmpegvideo_subtitle_info_s *subtitle_infos, unsigned int refresh_type);

protected:
	void resizeEvent(QResizeEvent *e);
	void paintEvent(QPaintEvent *event);
#ifdef DISPQT_USE_GDI_WIDGET
	QPaintEngine* paintEngine() const;
#endif

private:
	void videowidget_video_subtitle_render(QPainter *painter, ffmpegvideo_subtitle_info_s *subtitle_infos);
	void videowidget_ffmpeg_subtitle_store(ffmpegvideo_subtitle_info_s *subtitle_infos);
	void videowidget_ffmpeg_subtitle_close(void);
	void videowidget_ffmpeg_filtered_surface_update(bool immediate_paint);
	bool videowidget_ffmpeg_swsctx_calculate(int window_width, int window_height);

	bool videowidget_ffmpeg_surface_move_calculate(int move_x, int move_y);
	bool videowidget_ffmpeg_surface_zoom_calculate(int globalpos_x, int globalpos_y, int direction);
	bool videowidget_ffmpeg_handle_refresh(unsigned int refresh_type, bool doRepaint);

#ifdef DISPQT_USE_GDI_WIDGET
	void videowidget_video_gdi_render(struct dispqt_video_surface_info_s *video_surface_infos);
#else
	void videowidget_video_frame_render(QPainter *p);
#endif

	MainWindow *main_window;
	QWidget *parent_widget;
	FFMpegVideoSoftwareTransform *video_sw_transform;
	FFMpegVideoDecoderWorker *video_decoder_worker;
	ffmpegvideo_subtitle_info_s *sub_title_infos;
	dispqt_video_source_info_s video_source_infos;
	dispqt_video_surface_info_s video_surface_infos;
	QMutex mutex_video;
	QFont font_subtitle;
};

#endif // MPXPLAY_LINK_ORIGINAL_FFMPEG

class DispQtVideoPlayer : public QWidget
{
Q_OBJECT
public:
enum {VideoCtrlValue_FullScreen = 0, VideoCtrlValue_VideoWallAvailable, VideoCtrlValue_VideoWallEnabled, VideoCtrlValue_VideoWallARtype, VideoCtrlValue_Height, VideoCtrlValue_Speed};
	DispQtVideoPlayer(MainWindow *main_window = 0, QWidget *parent = 0);
	~DispQtVideoPlayer();
	void video_mediaplayer_open(bool lock_qmedia);
	void video_mediaplayer_close(void);
	void set_focus(void) { if(this->video_widget_context) this->video_widget_context->setFocus(Qt::MouseFocusReason); }
	void video_request_open(void *handle_ptr, char *utf8_filename);
	bool video_wait_for_open(void);
	void video_request_close(void *handle_ptr);
	bool video_wait_for_close(void);
	bool video_is_fullscreen(void) { return ((this->video_widget_context)? this->video_widget_context->isFullScreen() : false); }
	qint64 video_tell_ms(void);
	qint64 video_duration_ms(void);
	QWidget *video_get_widgetptr(void) { return this->video_widget_context; }
	int  video_get_value(unsigned int ctrl);
	void video_mouse_cursor_show(bool timed);
	void video_mouse_cursor_forbid_hide(bool forbid);
	void video_widget_init(void);
	void video_widget_control(unsigned long mainwin_control);
	//QString video_media_name(void) { return this->media_name; }
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	int video_ffmpeg_configcallback_create(unsigned int cfgcmd, mpxp_ptrsize_t cbk_func, mpxp_ptrsize_t passdata);
	bool video_ffmpeg_schedulerthread_suspend(void);
	void video_ffmpeg_schedulerthread_resume(void);
	void video_ffmpeg_framedisplay_wrapper(int video_index, unsigned int flags, void *videoout_frame, void *subtitle_infos, unsigned int refresh_type);
	DispQtVideoFrameFilter *video_frame_filter_transform_process;
#endif
	DispQtVideoFrameFilter *video_frame_filter_eq_process;

Q_SIGNALS:
	void signal_video_open(void);
	void signal_video_close(void);
	void signal_video_play(bool startflag);
	void signal_video_pause(void);
	void signal_video_switch_play(void);
	void signal_video_seek_ms(qint64 pos_ms);
	void signal_video_set_fullscreen(bool fs);
	void signal_video_switch_fullscreen(void);
	void signal_video_apply_fullscreen(bool fs); // 2nd step of switch_fullscreen
	void signal_video_check_windows(void);
	void signal_video_apply_windows(void);       // 3rd step of switch_fullscreen
	void signal_video_event_stop(void);
	void signal_video_set_speed(int speed1000);
	void signal_audio_set_volume(int vol);
	void signal_video_set_value(unsigned int ctrl, int val);
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	void signal_video_surface_resize_refresh(void);
	void signal_video_surface_filtered_trans_refresh(void);
	void signal_video_surface_filtered_eq_refresh(void);
	void signal_ffmpeg_configcallback_nonblock(unsigned int cfgcmd, mpxp_ptrsize_t cbk_func, mpxp_ptrsize_t passdata);
	void signal_ffmpeg_config_callback(unsigned int cfgcmd, mpxp_ptrsize_t cbk_func, mpxp_ptrsize_t passdata);
	void signal_ffmpeg_callback_call(int globalpos_x, int globalpos_y, unsigned long command, mpxp_ptrsize_t arg1, mpxp_ptrsize_t arg2);
#endif

private Q_SLOTS:
	void video_open(void);
	void video_close(void);
	void video_play(bool startflag);
	void video_pause(void);
	void video_switch_play(void) { if(this->is_playing) this->video_pause(); else this->video_play(false); }
	void video_stop(void *handle_ptr, bool close_media);
	void video_seek_ms(qint64 pos_ms);
	void video_set_fullscreen(bool fs);
	void video_switch_fullscreen(void);
	void video_apply_fullscreen(bool fs);
	void video_check_windows(void);
	void video_apply_windows(void);
	void video_handle_event_stop(void);
	void video_set_speed(int speed1000);
	void audio_set_volume(int vol);
	void video_set_value(unsigned int ctrl, int val);
#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG
	void video_surface_resize_refresh(void);
	void video_surface_filtered_trans_refresh(void);
	void video_surface_filtered_eq_refresh(void);
	void video_ffmpeg_configcallback_wrapper(unsigned int cfgcmd, mpxp_ptrsize_t cbk_func, mpxp_ptrsize_t passdata);
	void video_ffmpeg_config_callback(unsigned int cfgcmd, mpxp_ptrsize_t cbk_func, mpxp_ptrsize_t passdata);
	void video_ffmpeg_callback_call(int globalpos_x, int globalpos_y, unsigned long command, mpxp_ptrsize_t arg1, mpxp_ptrsize_t arg2);
#endif

#ifdef MPXPLAY_LINK_QTMEDIA
	void status_changed(QMediaPlayer::MediaStatus status);
	void display_error_message(QMediaPlayer::Error);
#endif

private:
#ifdef MPXPLAY_LINK_QTMEDIA
	void video_cursor_select(QMediaPlayer::MediaStatus status);
#endif

	MainWindow *main_window;
	QWidget *parent_widget;
	QHBoxLayout *video_layout;
	QString media_name;
#ifdef MPXPLAY_LINK_QTMEDIA
	QMediaPlayer *qt_media_player;
	QUrl media_url;
	QMediaContent media_content;
#endif
	DispQtVideoEventFilter *video_event_filter;
	QWidget *video_widget_context;
	Qt::WindowFlags video_wdg_nonfullscreen_flags;
	bool is_video_show, is_video_open, is_wait_for_open, is_playing;
	bool is_wait_for_close, is_video_closed, is_videowall_available, is_videowall_enabled;
	void *ptr_handle_curr, *ptr_handle_new, *ptr_handle_close; // for close identification
	QMutex mutex_player, mutex_qmedia;
};

#if defined(PDS_THREADS_POSIX_THREAD)
extern void *mpxplay_dispqt_video_decoder_worker_thread(void *decoder_data);
#else
extern unsigned int mpxplay_dispqt_video_decoder_worker_thread(void *decoder_data);
#endif

#endif // MOC_VIDEO_QT_H
