//**************************************************************************
//*                     This file is part of the                           *
//*                MMC - Mpxplay Multimedia Commander                      *
//*                   The source code of MMC is                            *
//*        (C) copyright 1998-2019 by PDSoft (Attila Padar)                *
//*                http://mpxplay.sourceforge.net                          *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: ffmpeg video decoding

//#define MPXPLAY_USE_DEBUGF 1
#define DISPQT_DEBUG_ERROR stderr
#define DISPQT_DEBUGOUT_WARNING stderr
#define DISPQT_DEBUGOUT_VTHREAD stdout
#define DISPQT_DEBUGOUT_CLEAR stdout
#define DISPQT_DEBUGOUT_VFCC NULL // stdout
#define DISPQT_DEBUGOUT_WV_MERGE NULL // stdout

#include "moc_video_qt.h"

#define INFFMPG_MUTEX_ONTIMER_TIMEOUT     10 // used on 10ms timer
#define INFFMPG_MUTEX_PACKETGET_TIMEOUT  100
#define INFFMPG_MUTEX_COMMAND_TIMEOUT   5000

#define INFFMPG_NB_LOST_FRAMES_LIMIT  500 // to switch of KeyframesOnly at videowall

#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG

#include <math.h>

static void videoworkdec_videowall_update_window_attribs(struct ffmpegvideo_callback_list_s *cbelem);

extern "C" {
 extern unsigned long mpxplay_config_video_audiovisualization_type;
}

struct dispqt_video_filter_format_s mpxplay_videoworkdec_seekpreview_format;
struct dispqt_video_filter_format_s mpxplay_videoworkdec_videowidget_format;

static void video_worker_videodecoder_context_process(struct ffmpegvideo_callback_list_s *cbelem, struct ffmpegvideo_decoder_s *fvdt, unsigned int *sleep_time, unsigned int *packet_flags)
{
	int streamtype_index = fvdt->streamtype_index, flags = MPXPLAY_PACKETLISTFLAG_CLEARBUF;
	mpxplay_packetlist_t *videoctx_pktelem = NULL;

	if((cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_CODECCTXINIT_FFMPEG, (mpxp_ptrsize_t)&videoctx_pktelem, streamtype_index, INFFMPG_MUTEX_PACKETGET_TIMEOUT) == MPXPLAY_ERROR_OK) && videoctx_pktelem)
	{
		fvdt->video_ctx = (AVCodecContext *)videoctx_pktelem->codec_ctx;
		//mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "video_decoder_worker_thread GET_CODECCTXINIT n:%d pd:%8.8X ctx:%8.8X", fvdt->streamtype_index, (unsigned int)cbelem->infile_passdata, (unsigned int)fvdt->video_ctx);
		if(!fvdt->video_pktlist_elem || (videoctx_pktelem->flags & MPXPLAY_PACKETLISTFLAG_CLEARBUF))
		{
			if(videoctx_pktelem->flags & MPXPLAY_PACKETLISTFLAG_CLEARBUF)
			{
				funcbit_disable(videoctx_pktelem->flags, MPXPLAY_PACKETLISTFLAG_CLEARBUF);
				fvdt->resetvideo = fvdt->resetfilter = fvdt->resetoutqueue = fvdt->resetcodec = true;
				mpxplay_debugf(DISPQT_DEBUGOUT_CLEAR, "CLEARBUF CTX s:%d", fvdt->streamtype_index);
			}
			if(videoctx_pktelem->flags & MPXPLAY_PACKETLISTFLAG_STILLPICT)
				funcbit_enable(*packet_flags, MPXPLAY_PACKETLISTFLAG_STILLPICT);
			if( (cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_PACKETQUEUE_CLEAR, streamtype_index, flags, INFFMPG_MUTEX_COMMAND_TIMEOUT) == 1)
			 || (fvdt->resetvideo && fvdt->video_pktlist_elem && (fvdt->video_pktlist_elem->codec_ctx != fvdt->video_ctx)) // TODO: check (wrong video_pktlist_elem->codec_ctx can crash the program)
			){
				if(fvdt->video_pktlist_elem){
					mpxplay_ffmpgdec_queuelistelem_clear(fvdt->video_pktlist_elem);
					fvdt->video_pktlist_elem = NULL;
				}
			}
			if(!fvdt->video_pktlist_elem)
				cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_PACKETLIST_ELEM, (mpxp_ptrsize_t)&fvdt->video_pktlist_elem, streamtype_index, INFFMPG_MUTEX_PACKETGET_TIMEOUT);
			if(fvdt->video_pktlist_elem)
			{
				if(funcbit_test(fvdt->video_pktlist_elem->flags, MPXPLAY_PACKETLISTFLAG_CLEARBUF))
				{
					fvdt->resetvideo = fvdt->resetfilter = fvdt->resetoutqueue = true;
					mpxplay_debugf(DISPQT_DEBUGOUT_CLEAR, "CLEARBUF PACKET s:%d dts:%lld", fvdt->streamtype_index, (fvdt->video_pktlist_elem->frame_pkt)? ((AVPacket *)fvdt->video_pktlist_elem->frame_pkt)->dts : 0);
				}
			}
			if(fvdt->resetvideo && funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO))
				funcbit_enable(fvdt->flags, (DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP | DISPQT_FFMPGVIDEOCALLBACKFLAG_DISPLAYONEFRAME));
			if(*sleep_time > MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT)
				*sleep_time = MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT;
		}
		else if(fvdt->video_ctx)
		{
			if(*sleep_time > MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT)
				*sleep_time = MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT;
		}
	}
}

static void video_worker_videodecoder_frameprocess_speedup_on_delay(struct ffmpegvideo_callback_list_s *cbelem, struct ffmpegvideo_decoder_s *fvdt, AVCodecContext *video_ctx)
{
	if( (fvdt->streamtype_index >= MPXPLAY_STREAMTYPEINDEX_VIDEOWALL) && (fvdt->streamtype_index != cbelem->selected_stream_index)
	 && funcbit_test(fvdt->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_ONLY) && (fvdt->video_decoder_lost_frames < INFFMPG_NB_LOST_FRAMES_LIMIT)
	 && !funcbit_test(fvdt->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_NO)
	){
		video_ctx->skip_loop_filter = AVDISCARD_ALL;
		video_ctx->skip_frame = AVDISCARD_NONKEY;
	}
	else if(fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_SEEKPREVIEW)
	{
		video_ctx->skip_loop_filter = AVDISCARD_ALL;
		video_ctx->skip_frame = AVDISCARD_DEFAULT;
	}
	else
	{
		video_ctx->skip_loop_filter = AVDISCARD_DEFAULT;
		video_ctx->skip_frame = AVDISCARD_DEFAULT;
		switch(cbelem->video_delay_type)
		{
			case DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SIGNIFICANT:
				video_ctx->skip_loop_filter = AVDISCARD_ALL;
				video_ctx->skip_frame = AVDISCARD_NONREF;
				//video_ctx->skip_frame = AVDISCARD_NONKEY; // FIXME: gives broken pictures restoring to AVDISCARD_DEFAULT (we should wait for a new keyframe before displaying frames)
				break;
			case DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_LARGE:
				video_ctx->skip_frame = AVDISCARD_NONREF;//AVDISCARD_NONINTRA;
				break;
			case DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_MIDDLE:
				video_ctx->skip_frame = AVDISCARD_NONREF;//AVDISCARD_BIDIR;
				break;
			case DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SMALL:
				break;
			default:
				video_ctx->skip_loop_filter = AVDISCARD_DEFAULT;
				break;
		}
		if(funcbit_test(fvdt->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_ONLY) && (fvdt->video_decoder_lost_frames >= INFFMPG_NB_LOST_FRAMES_LIMIT))
		{
			funcbit_enable(fvdt->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_NO);
		}
	}
}

static void video_worker_videodecoder_update_metadata_from_vframe(struct ffmpegvideo_callback_list_s *cbelem, struct ffmpegvideo_decoder_s *fvdt, AVFrame *vid_frame_decoded, unsigned int packet_flags)
{
	if(funcbit_test(packet_flags, MPXPLAY_PACKETLISTFLAG_STILLPICT) && vid_frame_decoded && (fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_VIDEO))
	{
		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_VIDFRAME_METADATA, (mpxp_ptrsize_t)vid_frame_decoded->metadata, 0, INFFMPG_MUTEX_COMMAND_TIMEOUT);
	}
}

static void videoworkdec_get_videoframe_timestamp(struct ffmpegvideo_callback_list_s *cbelem, struct ffmpegvideo_decoder_s *fvdt, AVFrame *vid_frame_filtered)
{
	cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_CORRECTEDVTIMESTAMP, (mpxp_ptrsize_t)vid_frame_filtered, (mpxp_ptrsize_t)&vid_frame_filtered->pts, INFFMPG_MUTEX_COMMAND_TIMEOUT);
}

static void videoworkdec_setget_videoframe_timestamp(struct ffmpegvideo_callback_list_s *cbelem, struct ffmpegvideo_decoder_s *fvdt, AVFrame *vid_frame_filtered)
{
	if(fvdt->resetvideo)
	{
		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_VIDEOTIMESTAMP, (mpxp_ptrsize_t)vid_frame_filtered, 0, INFFMPG_MUTEX_COMMAND_TIMEOUT);
	}
	else if(cbelem->video_delay_type >= DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SMALL)
	{
		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_VIDEOTIMESTAMP, (mpxp_ptrsize_t)vid_frame_filtered, 0, INFFMPG_MUTEX_ONTIMER_TIMEOUT);
	}
	videoworkdec_get_videoframe_timestamp(cbelem, fvdt, vid_frame_filtered);
}

#define MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_NONE    0
#define MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_FILTER  1
#define MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_ALL     2

static unsigned int video_worker_videodecoder_check_skip_frame(struct ffmpegvideo_callback_list_s *cbelem, struct ffmpegvideo_decoder_s *fvdt,
		unsigned int packet_flags, const mpxp_bool_t videowall_is_enabled)
{
	AVFrame *decoded_frame = fvdt->vid_frame_decoded;
	unsigned int skipframe_level = MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_ALL;

	if(!decoded_frame)
	{
		return skipframe_level;
	}

	if( !funcbit_test(packet_flags, MPXPLAY_PACKETLISTFLAG_STILLPICT)
	 || (decoded_frame->width > fvdt->video_last_valid_frame_w)
	 || (decoded_frame->height > fvdt->video_last_valid_frame_h)
	){
		if((fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_VIDEO) && videowall_is_enabled) // TODO: we should use the primary video stream for the selected_stream_index (currently it's a duplicated decoding)
		{
			skipframe_level = MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_FILTER;
		}
		else if((fvdt->streamtype_index != cbelem->selected_stream_index) && !fvdt->video_frame_filter_workerprocess)
		{
			skipframe_level = MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_FILTER;
		}
		else if(cbelem->video_delay_type == DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SIGNIFICANT)
		{
			fvdt->video_delay_counter_workdec_filter = MPXPLAY_VIDEO_WORKDEC_FILTER_RESTORE_DELAY;
			skipframe_level = MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_FILTER;
		}
		else if(fvdt->video_delay_counter_workdec_filter)
		{
			fvdt->video_delay_counter_workdec_filter--;
			skipframe_level = MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_FILTER;
		}
		else
		{
			skipframe_level = MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_NONE;
		}
	}
	fvdt->video_last_valid_frame_w = decoded_frame->width;
	fvdt->video_last_valid_frame_h = decoded_frame->height;

	return skipframe_level;
}

//-----------------------------------------------------------------------------------------------------------------------------------------

// check that videowall is enabled and read primary (currently played) window-id (streamtype_index) from the demultiplexer
static mpxp_bool_t videoworkdec_videowall_check(struct ffmpegvideo_callback_list_s *cbelem, unsigned int *streamindex_max)
{
	struct ffmpegvideo_decoder_s *fvdt = &cbelem->video_decoder_threads[MPXPLAY_STREAMTYPEINDEX_VIDEO];
	mpxp_int32_t new_sel_stream_index = MPXPLAY_STREAMTYPEINDEX_VIDEO;
	mpxp_bool_t videwall_is_enabled = cbelem->is_videowall_enabled;

	*streamindex_max = MPXPLAY_STREAMTYPEINDEX_NUM;
	if(videwall_is_enabled)
	{
		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_VIDEOWALLCURRSTI, (mpxp_ptrsize_t)&new_sel_stream_index, (mpxp_ptrsize_t)0, INFFMPG_MUTEX_PACKETGET_TIMEOUT);
		if(new_sel_stream_index >= MPXPLAY_STREAMTYPEINDEX_VIDEOWALL)
		{
			videoworkdec_videowall_update_window_attribs(cbelem);
			*streamindex_max = MPXPLAY_STREAMTYPEINDEX_MAX;
		}
		else
		{
			videwall_is_enabled = FALSE;
		}
	}

	if(new_sel_stream_index != cbelem->selected_stream_index)
	{
		cbelem->selected_stream_index = new_sel_stream_index;
		fvdt->resetfilter = true;
	}

	fvdt->videowall_window_width = mpxplay_videoworkdec_videowidget_format.width;
	fvdt->videowall_window_height = mpxplay_videoworkdec_videowidget_format.height;

	return videwall_is_enabled;
}

// update videowall stream number and calculate number of sub-windows (2x2, 3x3, 4x4, 5x5, 6x6)
static void videoworkdec_videowall_update_window_attribs(struct ffmpegvideo_callback_list_s *cbelem)
{
	struct dispqt_video_filter_format_s output_format = mpxplay_videoworkdec_videowidget_format;

	if((cbelem->videowall_count_streams) < 1 && (cbelem->videowall_nb_streams < 1))
		return;

	if(output_format.width < 128)
		output_format.width = 128;
	if(output_format.height < 128)
		output_format.height = 128;

	cbelem->videowall_output_format = output_format;
	cbelem->videowall_output_artype = cbelem->video_wall_artype;

	if(cbelem->videowall_count_streams > cbelem->videowall_nb_streams) // number of visible windows (decoded video streams) is increased
		cbelem->videowall_nb_streams = cbelem->videowall_count_streams;
	//else if(cbelem->videowall_count_streams < cbelem->videowall_nb_streams)
	// TODO: timer/counter to reduce number of windows at the case of lost video stream

	cbelem->videowall_nb_windows_x = (int)sqrt((double)cbelem->videowall_nb_streams); // eg: 16 streams -> 4 x 4 wall
	if(cbelem->videowall_nb_windows_x <= 0)
		cbelem->videowall_nb_windows_x = 1;

	if((output_format.width * 9 / output_format.height / 15) > cbelem->videowall_nb_windows_x) // vertical narrow layout (N x 1)
	{
		cbelem->videowall_nb_windows_x = cbelem->videowall_nb_streams;
		cbelem->videowall_nb_windows_y = 1;
	}
	else if((output_format.height * 15 / output_format.width / 10) > cbelem->videowall_nb_windows_x) // horizontal narrow layout (1 x N)
	{
		cbelem->videowall_nb_windows_x = 1;
		cbelem->videowall_nb_windows_y = cbelem->videowall_nb_streams;
	}
	else
	{
		cbelem->videowall_nb_windows_y = cbelem->videowall_nb_windows_x;
		if((cbelem->videowall_nb_windows_x * cbelem->videowall_nb_windows_y) < cbelem->videowall_nb_streams) // nb_streams is not equal with 3x3, 4x4, etc
		{
			if(output_format.width > output_format.height) // horizontal layout (3x2, 4x3)
				cbelem->videowall_nb_windows_x++;
			else                                           // vertical layout (2x3, 3x4)
				cbelem->videowall_nb_windows_y++;
		}
	}

	// calculate the window size of a video stream (sub-picture)
	output_format.width = output_format.width / cbelem->videowall_nb_windows_x;
	output_format.width = output_format.width & ~7; // TODO: avcodec_align_dimensions()
	output_format.height = output_format.height / (((cbelem->videowall_output_artype == DISPQT_VIDEOFULLSCREENARTYPE_FILL) || (cbelem->videowall_nb_windows_y > cbelem->videowall_nb_windows_x))? cbelem->videowall_nb_windows_y : cbelem->videowall_nb_windows_x);
	output_format.height = output_format.height & ~3;

	if((output_format.width != cbelem->videowall_subwindow_format.width) || (output_format.height != cbelem->videowall_subwindow_format.height) || (output_format.pixelformat != cbelem->videowall_subwindow_format.pixelformat))
	{
		struct ffmpegvideo_decoder_s *fvdt = &cbelem->video_decoder_threads[MPXPLAY_STREAMTYPEINDEX_VIDEOWALL];
		int i;
		for(i = MPXPLAY_STREAMTYPEINDEX_VIDEOWALL; i < MPXPLAY_STREAMTYPEINDEX_MAX; i++, fvdt++)
		{
			if(fvdt->video_frame_decoded_preprocessed)
				av_frame_free(&fvdt->video_frame_decoded_preprocessed);
			funcbit_disable(fvdt->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_ONLY);
		}
		cbelem->videowall_subwindow_format = output_format;
	}

	cbelem->videowall_count_streams = 0;
}

// copy decoded and resized sub-video frames into one screen-size video frame   (FIXME: this is a slow software solution)
static void videoworkdec_videowall_merge_windows(struct ffmpegvideo_callback_list_s *cbelem)
{
	struct ffmpegvideo_decoder_s *fvdt = &cbelem->video_decoder_threads[MPXPLAY_STREAMTYPEINDEX_VIDEO];
	const unsigned int packet_flags = (fvdt->resetvideo || fvdt->resetoutqueue)? MPXPLAY_PACKETLISTFLAG_CLEARBUF : 0;
	AVFrame *vid_frame_merged, *vid_frame_primaryvideo;
	const int bytes_per_pixel = 4; // TODO: non-RGB output
	int i, w, linebytes_in, linebytes_out, vertical_shift;
	int videowall_nb_windows_max;

	cbelem->videowall_count_streams = 0;

	vid_frame_primaryvideo = fvdt->video_frame_decoded_preprocessed;
	if(!vid_frame_primaryvideo) // we synchronize to the primary decoded video frame
		return;

	if((fvdt->videoframe_queue.nb_packets >= DISPQT_FFMPGVIDEO_MAX_QUEUED_VIDEOFRAME) && !fvdt->resetvideo)
		return;

	fvdt->resetvideo = fvdt->resetoutqueue = fvdt->resetcodec = false;

	cbelem->videowall_sync_timestamp = vid_frame_primaryvideo->pts;

	vid_frame_merged = av_frame_alloc();
	if(!vid_frame_merged)
		return;

	videowall_nb_windows_max = cbelem->videowall_nb_windows_x * cbelem->videowall_nb_windows_y;

	if( (cbelem->videowall_output_artype != DISPQT_VIDEOFULLSCREENARTYPE_FILL)
	 && (cbelem->videowall_nb_windows_x >= cbelem->videowall_nb_windows_y)
	 && (cbelem->videowall_nb_streams <= (cbelem->videowall_nb_windows_x * (cbelem->videowall_nb_windows_x - 1))) // eg. 3 x 2 windows
	){   // then reduce output frame height to put it vertically centered
		vid_frame_merged->height = cbelem->videowall_output_format.height * (cbelem->videowall_nb_windows_x - 1) / cbelem->videowall_nb_windows_x;
		vertical_shift = (cbelem->videowall_output_format.height - vid_frame_merged->height) / 2; // FIXME: bullshit
	}
	else
	{
		vid_frame_merged->height = cbelem->videowall_output_format.height;
		vertical_shift = 0;
	}
	vid_frame_merged->width = cbelem->videowall_output_format.width;
	vid_frame_merged->format = cbelem->videowall_output_format.pixelformat;
	vid_frame_merged->sample_aspect_ratio.num = vid_frame_merged->sample_aspect_ratio.den = 0;

	if(av_frame_get_buffer(vid_frame_merged, 0) != 0)
	{
		av_frame_free(&vid_frame_merged);
		return;
	}

	linebytes_in = cbelem->videowall_subwindow_format.width * bytes_per_pixel;
	linebytes_out = vid_frame_merged->linesize[0];

	//if((linebytes_out % linebytes_in) || (cbelem->videowall_nb_streams < videowall_nb_windows_max)) // TODO:
	{
		pds_memset(vid_frame_merged->data[0], 0, vid_frame_merged->height * linebytes_out);
	}

	fvdt = &cbelem->video_decoder_threads[MPXPLAY_STREAMTYPEINDEX_VIDEOWALL];
	for(i = MPXPLAY_STREAMTYPEINDEX_VIDEOWALL, w = 0; i < MPXPLAY_STREAMTYPEINDEX_MAX; i++, fvdt++)
	{
		AVFrame *vid_frame_window = fvdt->video_frame_decoded_preprocessed;
		if(vid_frame_window)
		{
			if((w < videowall_nb_windows_max) && (vid_frame_window->linesize[0] >= linebytes_in) && (vid_frame_window->format == cbelem->videowall_subwindow_format.pixelformat))
			{
				const int destpos = (w / cbelem->videowall_nb_windows_x) * linebytes_out * cbelem->videowall_subwindow_format.height + (w % cbelem->videowall_nb_windows_x) * linebytes_in;
				uint8_t *destptr = vid_frame_merged->data[0] + destpos;
				uint8_t *srcptr =  vid_frame_window->data[0];
				int j = vid_frame_window->height;

				do{
					pds_memcpy(destptr, srcptr, linebytes_in);
					srcptr+= linebytes_in;
					destptr += linebytes_out;
				}while(--j);

				fvdt->videowall_window_pos_x = (w % cbelem->videowall_nb_windows_x) * cbelem->videowall_subwindow_format.width;
				fvdt->videowall_window_pos_y = (w / cbelem->videowall_nb_windows_x) * cbelem->videowall_subwindow_format.height + vertical_shift;
				fvdt->videowall_window_width = vid_frame_window->width;
				fvdt->videowall_window_height = vid_frame_window->height;
				if(!w)
				{
					mpxplay_debugf(DISPQT_DEBUGOUT_WV_MERGE, "video_merge w:%d h:%d ow:%d oh:%d", fvdt->videowall_window_width, fvdt->videowall_window_height,
							cbelem->videowall_output_format.width / cbelem->videowall_nb_windows_x, cbelem->videowall_output_format.height / cbelem->videowall_nb_windows_y);
				}
			}
			else
			{
				fvdt->videowall_window_width = 0;
			}
			w++;
		}
		else
		{
			fvdt->videowall_window_width = 0;
		}
	}
	cbelem->videowall_count_streams = w;

	vid_frame_merged->pts = vid_frame_primaryvideo->pts;
	vid_frame_merged->pkt_pts = vid_frame_primaryvideo->pkt_pts; // TODO: remove
	vid_frame_merged->pkt_dts = vid_frame_primaryvideo->pkt_dts;
	vid_frame_merged->best_effort_timestamp = vid_frame_primaryvideo->best_effort_timestamp;
	vid_frame_merged->pkt_duration = vid_frame_primaryvideo->pkt_duration;

	fvdt = &cbelem->video_decoder_threads[MPXPLAY_STREAMTYPEINDEX_VIDEO]; // TODO: we should use the primary video stream for the selected_stream_index (currently it's a duplicated decoding)
	av_frame_free(&fvdt->video_frame_decoded_preprocessed);
	mpxplay_ffmpgdec_packetqueue_put(&fvdt->videoframe_queue, MPXPLAY_PACKETTYPE_AVFRAME_VIDEO, packet_flags, NULL, vid_frame_merged);

	mpxplay_debugf(DISPQT_DEBUGOUT_WV_MERGE, "video_merge DONE psi:%d w:%d pts:%lld", cbelem->selected_stream_index, cbelem->videowall_count_streams, vid_frame_merged->pts);
}

//-----------------------------------------------------------------------------------------------------------------------------------------

static int video_worker_videodecoder_filter_process(struct ffmpegvideo_callback_list_s *cbelem, struct ffmpegvideo_decoder_s *fvdt, AVFrame **vid_frame_filtered,
		unsigned int packet_flags, const mpxp_bool_t videowall_is_enabled)
{
	DispQtVideoFrameFilter *video_frame_filter = ((fvdt->streamtype_index == cbelem->selected_stream_index) && !fvdt->video_frame_filter_workerprocess)? cbelem->video_frame_filter_workerprocess : fvdt->video_frame_filter_workerprocess;
	int filter_retVal = DISPQT_VIDEOFILTER_APPLYRETBIT_INPUTFRAME_PROCESSED;
	unsigned int skipframe_level;

	skipframe_level = video_worker_videodecoder_check_skip_frame(cbelem, fvdt, packet_flags, videowall_is_enabled);
	if(skipframe_level == MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_ALL)
	{
		av_frame_free(&fvdt->vid_frame_decoded);
		*vid_frame_filtered = NULL;
		//fvdt->resetvideo = false;
	}
	else if((skipframe_level == MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_NONE) && video_frame_filter)
	{
		filter_retVal = video_frame_filter->video_filter_apply(fvdt->vid_frame_decoded, vid_frame_filtered, AV_PIX_FMT_NONE);
		if(fvdt->vid_frame_decoded)
		{
			if(funcbit_test(filter_retVal, DISPQT_VIDEOFILTER_APPLYRETBIT_INPUTFRAME_PROCESSED))
			{
				av_frame_free(&fvdt->vid_frame_decoded);
			}
			else if(!funcbit_test(filter_retVal, DISPQT_VIDEOFILTER_APPLYRETBIT_OUTPUTFRAME_PRODUCED))
			{
				filter_retVal = DISPQT_VIDEOFILTER_APPLYRETBIT_INPUTFRAME_PROCESSED;
				*vid_frame_filtered = fvdt->vid_frame_decoded;
				fvdt->vid_frame_decoded = NULL;
			}
		}
	}
	else
	{
		*vid_frame_filtered = fvdt->vid_frame_decoded;
		fvdt->vid_frame_decoded = NULL;
	}

	// convert video frame to a smaller one and to a target pixelformat (seekpreview and videowall)
	if((filter_retVal > DISPQT_VIDEOFILTER_APPLYRETBIT_NOFILTERAPPLY) && *vid_frame_filtered && (fvdt->streamtype_index >= MPXPLAY_STREAMTYPEINDEX_SEEKPREVIEW))
	{
		AVFrame *vid_frame_input = *vid_frame_filtered;
		struct dispqt_video_filter_format_s output_format;

		*vid_frame_filtered = NULL;

		if(fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_SEEKPREVIEW)
		{
			output_format = mpxplay_videoworkdec_seekpreview_format;

			if(output_format.width == 0) // then use video aspect ratio
			{
				int ar_num = (vid_frame_input->sample_aspect_ratio.num > 0)? vid_frame_input->sample_aspect_ratio.num : 1;
				int ar_den = (vid_frame_input->sample_aspect_ratio.den > 0)? vid_frame_input->sample_aspect_ratio.den : 1;
				output_format.width = vid_frame_input->width * output_format.height / vid_frame_input->height * ar_num / ar_den;
				output_format.width = ((output_format.width + 4) & ~7); // TODO: avcodec_align_dimensions()
			}
			else if(output_format.width == 1) // then use pixel aspect ratio
			{
				output_format.width = vid_frame_input->width * output_format.height / vid_frame_input->height;
				output_format.width = ((output_format.width + 4) & ~7); // TODO: avcodec_align_dimensions()
			}
		}
		else // MPXPLAY_STREAMTYPEINDEX_VIDEOWALL
		{
			output_format = cbelem->videowall_subwindow_format;
		}

		if(mpxplay_dispqt_videotrans_ffmpeg_swsctx_conv_frame(&fvdt->frame_conv_infos, vid_frame_input, &output_format))
		{
			*vid_frame_filtered = av_frame_clone(fvdt->frame_conv_infos.ffmpeg_sws_frame);
			(*vid_frame_filtered)->pts = vid_frame_input->pts;
			(*vid_frame_filtered)->pkt_dts = vid_frame_input->pkt_dts;
			(*vid_frame_filtered)->best_effort_timestamp = vid_frame_input->best_effort_timestamp;
			(*vid_frame_filtered)->pkt_duration = vid_frame_input->pkt_duration;
		}
		av_frame_free(&vid_frame_input);
	}

	return filter_retVal;
}

#ifdef MPXPLAY_LINK_DXVA
/* retrieve data from GPU to CPU */
static int video_worker_videodecoder_gpu_retrieve_data(AVFrame *frame)
{
	AVFrame *tmp_frame = av_frame_alloc();
	int ret;

	ret = av_hwframe_transfer_data(tmp_frame, frame, 0);
	if (ret < 0)
		return ret;

	ret = av_frame_copy_props(tmp_frame, frame);
	if(ret < 0)
	{
		av_frame_unref(tmp_frame);
		return ret;
	}

	av_frame_unref(frame);
	av_frame_move_ref(frame, tmp_frame);

	return 0;
}
#endif

static void video_worker_videodecoder_frame_process(struct ffmpegvideo_callback_list_s *cbelem, struct ffmpegvideo_decoder_s *fvdt, AVCodecContext *video_ctx,
		unsigned int *sleep_time, unsigned int packet_flags, const mpxp_bool_t videowall_is_enabled)
{
	int retcode_video_receive = -1, retcode_video_send = -1, retry_videodec = 3, packet_used = 0;
	AVFrame *vid_frame_filtered = NULL;

	if(fvdt->resetfilter)
	{
		av_frame_free(&fvdt->vid_frame_decoded);
		av_frame_free(&fvdt->video_frame_decoded_preprocessed);
		if(cbelem->video_frame_filter_workerprocess && (fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_VIDEO))
			cbelem->video_frame_filter_workerprocess->video_filter_close();  // TODO: other way to clear?
		if(fvdt->video_frame_filter_workerprocess)
			fvdt->video_frame_filter_workerprocess->video_filter_close();    //
		funcbit_disable(fvdt->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_ONLY);
		fvdt->resetfilter = false;
	}

	if(!video_ctx)
		return;

	if(!videowall_is_enabled || (fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_SEEKPREVIEW))
	{
		int max_queued_videoframe = (fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_VIDEO)? DISPQT_FFMPGVIDEO_MAX_QUEUED_VIDEOFRAME : ((fvdt->streamtype_index != MPXPLAY_STREAMTYPEINDEX_SEEKPREVIEW)? 1 : 0);
		if((fvdt->videoframe_queue.nb_packets >= max_queued_videoframe) && !fvdt->resetvideo)
			return;
	}
	else if(fvdt->video_frame_decoded_preprocessed && !fvdt->resetvideo
		 && ( (fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_VIDEO)
		   || ((fvdt->streamtype_index == cbelem->selected_stream_index) && (fvdt->video_frame_decoded_preprocessed->pts > cbelem->videowall_sync_timestamp))) // don't let the main VIDEO stream to run away (syncing with the primary/selected videowall stream)
	){
		return;
	}

	do{
		if(!fvdt->vid_frame_decoded)
		{
			fvdt->vid_frame_decoded = av_frame_alloc();
			if(!fvdt->vid_frame_decoded)
				break;

			video_worker_videodecoder_frameprocess_speedup_on_delay(cbelem, fvdt, video_ctx);

			retcode_video_receive = avcodec_receive_frame(video_ctx, fvdt->vid_frame_decoded);

			if(retcode_video_receive == 0)
			{
				fvdt->video_decoder_lost_frames = 0;
				if(funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_ONLY))
					funcbit_enable(fvdt->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_ONLY); // we switch on this feature only after the first frame received (for a faster initial displaying of videowall)
				else
					funcbit_disable(fvdt->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_VW_KEYFRAMES_ONLY);
				if((fvdt->streamtype_index >= MPXPLAY_STREAMTYPEINDEX_VIDEOWALL) && (fvdt->streamtype_index != cbelem->selected_stream_index))
				{
					mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "video_worker_videodecoder_frame_process FRAME i:%d r:%d", fvdt->streamtype_index, retcode_video_receive);
				}
			}

			video_worker_videodecoder_update_metadata_from_vframe(cbelem, fvdt, fvdt->vid_frame_decoded, packet_flags);
#if 0		// doesn't work
			if((retcode_video_receive < 0) || (cbelem->vid_frame_decoded->flags & AV_FRAME_FLAG_CORRUPT)) // || (vid_frame->decode_error_flags & (FF_DECODE_ERROR_INVALID_BITSTREAM | FF_DECODE_ERROR_MISSING_REFERENCE)))
			{
				if(retcode_video_receive != AVERROR(EAGAIN))
				{
					mpxplay_debugf(DISPQT_DEBUGOUT_WARNING, "avcodec_receive_frame ERROR %8.8X r:%d f:%d df:%d d:%8.8X ls:%d ", (unsigned int)cbelem,
							retcode_video_receive, cbelem->vid_frame_decoded->flags, cbelem->vid_frame_decoded->decode_error_flags, (unsigned int)cbelem->vid_frame_decoded->data[0], cbelem->vid_frame_decoded->linesize[0]);
				}
			}
#endif
#ifdef MPXPLAY_LINK_DXVA
			if( (retcode_video_receive == 0)
			 && ((fvdt->vid_frame_decoded->format == AV_PIX_FMT_DXVA2_VLD) || (fvdt->vid_frame_decoded->format == AV_PIX_FMT_D3D11)  || (fvdt->vid_frame_decoded->format == AV_PIX_FMT_D3D11VA_VLD))
			){
				retcode_video_receive = video_worker_videodecoder_gpu_retrieve_data(fvdt->vid_frame_decoded);
			}
#endif
			if((retcode_video_receive == 0) && (!fvdt->vid_frame_decoded->data[0] || (fvdt->vid_frame_decoded->linesize[0] <= 0) || (fvdt->vid_frame_decoded->flags & AV_FRAME_FLAG_DISCARD)))
				retcode_video_receive = -1;
			if(retcode_video_receive < 0)
				av_frame_free(&fvdt->vid_frame_decoded);
		}

		if(video_worker_videodecoder_filter_process(cbelem, fvdt, &vid_frame_filtered, packet_flags, videowall_is_enabled) < 0)
		{
			break;
		}

		if(vid_frame_filtered || packet_used || !fvdt->video_pktlist_elem)
		{
			break;
		}

		AVCodecContext *packet_vcodec_ctx = (AVCodecContext *)fvdt->video_pktlist_elem->codec_ctx;
		if(!packet_vcodec_ctx || !fvdt->video_pktlist_elem->frame_pkt || (fvdt->video_pktlist_elem->packet_type != MPXPLAY_PACKETTYPE_AVPACKET))
		{
			packet_used = 1;
			break;
		}
		if((packet_vcodec_ctx->codec_type != AVMEDIA_TYPE_VIDEO) || (packet_vcodec_ctx->codec_id == AV_CODEC_ID_NONE) || !avcodec_is_open(packet_vcodec_ctx) || !packet_vcodec_ctx->codec) // codec is not initialized (avcodec_open2 not called)
		{
			fvdt->videoctx_init_error_count++;
			if(fvdt->videoctx_init_error_count > INFFMPEG_VIDEO_INITERROR_LIMIT) // we drop the packet (containing the incorrect ctx) after a time
				packet_used = 1;
			break;
		}

		retcode_video_send = avcodec_send_packet(packet_vcodec_ctx, (AVPacket *)fvdt->video_pktlist_elem->frame_pkt);
		if(retcode_video_send == 0)
		{
			video_ctx = packet_vcodec_ctx;
			packet_used = 1;
			fvdt->video_decoder_lost_frames++;
//			if((fvdt->streamtype_index >= MPXPLAY_STREAMTYPEINDEX_VIDEOWALL) && (fvdt->streamtype_index != cbelem->selected_stream_index))
//			{
//				mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "video_worker_videodecoder_frame_process PACKET i:%d r:%d", fvdt->streamtype_index, retcode_video_send);
//			}
		}
		else if(retcode_video_send < 100)
		{
			AVPacket *av_pkt = (AVPacket *)fvdt->video_pktlist_elem->frame_pkt;
			mpxplay_debugf(DISPQT_DEBUG_ERROR, "avcodec_send_packet VIDEO failed ret:%d st:%d si:%d dts:%lld pctx:%8.8X vctx:%8.8X open:%d",
					retcode_video_send, fvdt->streamtype_index, av_pkt->stream_index, av_pkt->dts, packet_vcodec_ctx, video_ctx, avcodec_is_open(packet_vcodec_ctx));
			avcodec_flush_buffers(packet_vcodec_ctx); // TODO: check
			packet_used = 1;                          //
		}
	}while(--retry_videodec);

	if(packet_used || !retry_videodec) // clear packet if it's sent or seems to be unusefull
	{
		mpxplay_ffmpgdec_queuelistelem_clear(fvdt->video_pktlist_elem);
		fvdt->video_pktlist_elem = NULL;
		fvdt->videoctx_init_error_count = 0;
		*sleep_time = 0;
	}
	if(vid_frame_filtered)
	{
		if(vid_frame_filtered->width <= 0)
			vid_frame_filtered->width = video_ctx->width;
		if(vid_frame_filtered->height <= 0)
			vid_frame_filtered->height = video_ctx->height;
		if(vid_frame_filtered->format < 0)
			vid_frame_filtered->format = ((video_ctx->pix_fmt > AV_PIX_FMT_NONE)? video_ctx->pix_fmt : AV_PIX_FMT_YUV420P); // FIXME: default AV_PIX_FMT_YUV420P
		if(funcbit_test(packet_flags, MPXPLAY_PACKETLISTFLAG_STILLPICT)) // TODO: check: still pictures always use pixel aspect ratio override
		{
			vid_frame_filtered->sample_aspect_ratio.num = vid_frame_filtered->sample_aspect_ratio.den = 1;
		}
		else if((vid_frame_filtered->sample_aspect_ratio.num < video_ctx->sample_aspect_ratio.num) || (vid_frame_filtered->sample_aspect_ratio.den < video_ctx->sample_aspect_ratio.den))
		{
			vid_frame_filtered->sample_aspect_ratio.num = video_ctx->sample_aspect_ratio.num;
			vid_frame_filtered->sample_aspect_ratio.den = video_ctx->sample_aspect_ratio.den;
		}

		if(fvdt->resetcodec)
		{
			funcbit_enable(packet_flags, MPXPLAY_PACKETLISTFLAG_RESET);
		}
		if(fvdt->resetvideo)
		{
			funcbit_enable(packet_flags, MPXPLAY_PACKETLISTFLAG_CLEARBUF);
		}

		if(fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_VIDEO)
		{
			videoworkdec_setget_videoframe_timestamp(cbelem, fvdt, vid_frame_filtered);
		}
		else if(fvdt->streamtype_index == cbelem->selected_stream_index)
		{
			videoworkdec_get_videoframe_timestamp(cbelem, fvdt, vid_frame_filtered);
		}

		if(fvdt->video_frame_decoded_preprocessed)
		{
			av_frame_free(&fvdt->video_frame_decoded_preprocessed);
		}

		if(!videowall_is_enabled || (fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_SEEKPREVIEW))
		{
			mpxplay_ffmpgdec_packetqueue_put(&fvdt->videoframe_queue, MPXPLAY_PACKETTYPE_AVFRAME_VIDEO, packet_flags, video_ctx, vid_frame_filtered);
		}
		else // save frame for videowall windows merge
		{
			fvdt->video_frame_decoded_preprocessed = vid_frame_filtered;
		}

		if(!videowall_is_enabled || (fvdt->streamtype_index != MPXPLAY_STREAMTYPEINDEX_VIDEO))
		{
			fvdt->resetvideo = fvdt->resetoutqueue = fvdt->resetcodec = false;
		}

		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_CODECCTX_CONSOLIDATE, (mpxp_ptrsize_t)video_ctx, fvdt->streamtype_index, INFFMPG_MUTEX_COMMAND_TIMEOUT);

		*sleep_time = 0;

//		if((fvdt->streamtype_index == cbelem->selected_stream_index) || (fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_VIDEO) )
//		{
//			mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "video_decoder DONE s:%d w:%3d h:%3d pts:%lld skip:%d nb:%d pfl:%8.8X pn:%d", fvdt->streamtype_index, vid_frame_filtered->width, vid_frame_filtered->height,
//				 vid_frame_filtered->pts, video_ctx->skip_frame, fvdt->videoframe_queue.nb_packets, packet_flags, vid_frame_filtered->display_picture_number);
//		}
	}
}

static void video_worker_audiovisualization_process(struct ffmpegvideo_callback_list_s *cbelem, struct ffmpegvideo_decoder_s *fvdt, bool *resetvideo, unsigned int *sleep_time, unsigned int packet_flags)
{
	int streamtype_index = MPXPLAY_STREAMTYPEINDEX_VIDEO;
	bool audio_input_processed = true;
	if(!fvdt->video_pktlist_elem)
		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_PACKETLIST_ELEM, (mpxp_ptrsize_t)&fvdt->video_pktlist_elem, streamtype_index, INFFMPG_MUTEX_PACKETGET_TIMEOUT);
	if(fvdt->video_pktlist_elem)
	{
		AVFrame *audio_frame_decoded = (AVFrame *)fvdt->video_pktlist_elem->frame_pkt;
		if((fvdt->video_pktlist_elem->flags & MPXPLAY_PACKETLISTFLAG_CLEARBUF) || (fvdt->videoframe_queue.nb_packets > DISPQT_FFMPGVIDEO_MAX_DELAY_VISVIDEOFRAME))
		{
			funcbit_disable(fvdt->video_pktlist_elem->flags, MPXPLAY_PACKETLISTFLAG_CLEARBUF);
			*resetvideo = true;
		}
		//mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "audio visualization pe:%8.8X afd:%8.8X n:%d", (int)cbelem->video_pktlist_elem, (int)audio_frame_decoded, cbelem->videoframe_queue.nb_packets);
		if(audio_frame_decoded && (fvdt->streamtype_index == MPXPLAY_STREAMTYPEINDEX_VIDEO) && (fvdt->video_pktlist_elem->packet_type == MPXPLAY_PACKETTYPE_AVFRAME_AUDIO) && (fvdt->videoframe_queue.nb_packets < DISPQT_FFMPGVIDEO_MAX_QUEUED_VISVIDEOFRAME) && cbelem->video_frame_filter_workerprocess)
		{
			AVFrame *vid_frame_visual = NULL;
			int retcode = cbelem->video_frame_filter_workerprocess->video_filter_apply(audio_frame_decoded, &vid_frame_visual, DISPQT_VIDEO_FILTERPROCESS_AVFORMAT);
			if(vid_frame_visual)
			{
				if(*resetvideo)
				{
					cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_VIDEOTIMESTAMP, (mpxp_ptrsize_t)vid_frame_visual, 0, INFFMPG_MUTEX_COMMAND_TIMEOUT);
					cbelem->video_frame_filter_workerprocess->video_filter_close();  // TODO: other way to reset?
					funcbit_enable(packet_flags, MPXPLAY_PACKETLISTFLAG_CLEARBUF);
					funcbit_enable(fvdt->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP);
					*resetvideo = false;
					mpxplay_debugf(DISPQT_DEBUGOUT_CLEAR, "audio visualization RESET n:%d w:%d h:%d pts:%lld", fvdt->videoframe_queue.nb_packets, vid_frame_visual->width, vid_frame_visual->height, vid_frame_visual->pts);
				}
				// FIXME: this is a workaround for large audio blocks + visual plugins
				else
				{
#if 1
					mpxp_int64_t diff = vid_frame_visual->pts - fvdt->video_dec_last_timestamp;
					mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "vid_frame_visual pts:%lld DIFF:%lld", vid_frame_visual->pts, diff);
					if((diff > -(AV_TIME_BASE / 24)) && (diff < (AV_TIME_BASE / 50))) // visual plugin shall send with 25 fps
					{
						vid_frame_visual->pts = fvdt->video_dec_last_timestamp + (AV_TIME_BASE / 26);
					}
#endif
					if(!funcbit_test(retcode, DISPQT_VIDEOFILTER_APPLYRETBIT_INPUTFRAME_PROCESSED))
					{
						audio_input_processed = false;
					}
				}
				fvdt->video_dec_last_timestamp = vid_frame_visual->pts;
				//funcbit_enable(packet_flags, MPXPLAY_PACKETLISTFLAG_IGNORE_TS);
				//funcbit_enable(fvdt->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP); // for testing only
				funcbit_enable(vid_frame_visual->flags, (DISPQT_AVFRAME_FLAG_NOSKIPFRAME | DISPQT_AVFRAME_FLAG_AR_WINDOW));
				if(mpxplay_ffmpgdec_packetqueue_put(&fvdt->videoframe_queue, MPXPLAY_PACKETTYPE_AVFRAME_VIDEO, packet_flags, NULL, vid_frame_visual) < 0)
				{
					av_frame_free(&vid_frame_visual);
				}
				//mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "mpxplay_ffmpgdec_packetqueue_put DONE n:%d w:%d h:%d pts:%lld", cbelem->videoframe_queue.nb_packets, vid_frame_visual->width, vid_frame_visual->height, vid_frame_visual->pts);
				if(*sleep_time > MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT)
					*sleep_time = MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT;
			}
			else
			{
				*sleep_time = 0;
			}
		}

		if(audio_input_processed)
		{
			mpxplay_ffmpgdec_queuelistelem_clear(fvdt->video_pktlist_elem);
			fvdt->video_pktlist_elem = NULL;
		}
	}
}

#if defined(PDS_THREADS_POSIX_THREAD)
void *mpxplay_dispqt_video_decoder_worker_thread(void *decoder_data)
#else
unsigned int mpxplay_dispqt_video_decoder_worker_thread(void *decoder_data)
#endif
{
	struct ffmpegvideo_callback_list_s *cbelem = (struct ffmpegvideo_callback_list_s *)decoder_data;
	int cberror_count = 0, stream_index;

	do{
		unsigned int streamindex_max, packet_flags, sleep_time = MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_DEFAULT;
		mpxp_bool_t videwall_is_enabled;

		if(!cbelem || !cbelem->infile_callback)
		{
			if(++cberror_count < MPXPLAY_VIDEO_WORKER_THREAD_CRITICAL_ERROR_COUNT)
			{
				pds_threads_sleep(sleep_time);
				continue;
			}
			mpxplay_debugf(DISPQT_DEBUG_ERROR, "video_decoder_worker_thread ERROR cbelem");
			break;
		}

		videwall_is_enabled = videoworkdec_videowall_check(cbelem, &streamindex_max);

		cberror_count = 0;
		for(stream_index = MPXPLAY_STREAMTYPEINDEX_VIDEO; stream_index < streamindex_max; stream_index++)
		{
			struct ffmpegvideo_decoder_s *fvdt = &cbelem->video_decoder_threads[stream_index];

			if(stream_index == MPXPLAY_STREAMTYPEINDEX_SUBTITLE)
				continue;

			fvdt->video_ctx = NULL;
			packet_flags = 0;
			video_worker_videodecoder_context_process(cbelem, fvdt, &sleep_time, &packet_flags);

			if(funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO))
			{
				if(videwall_is_enabled && (stream_index == MPXPLAY_STREAMTYPEINDEX_VIDEO))
				{
					if(fvdt->video_ctx)
					{
						cbelem->is_videowall_disabled_by_decoder = FALSE;
					}
					else
					{
						cbelem->is_videowall_disabled_by_decoder = TRUE;
						videwall_is_enabled = FALSE;
						streamindex_max = MPXPLAY_STREAMTYPEINDEX_NUM;
					}
				}
				if(fvdt->video_ctx) // TODO: extra switch to show audio visualization for video files too
				{
					video_worker_videodecoder_frame_process(cbelem, fvdt, fvdt->video_ctx, &sleep_time, packet_flags, videwall_is_enabled);
					//mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "video_decoder_worker_thread videodecoder_frame_process END st:%d ctx:%8.8X", stream_index, (unsigned int)fvdt->video_ctx);
				}
				else if((mpxplay_config_video_audiovisualization_type != DISPQT_AUDIOVISUALIZATIONTYPE_NONE) && (stream_index == MPXPLAY_STREAMTYPEINDEX_VIDEO))
				{
					video_worker_audiovisualization_process(cbelem, fvdt, &fvdt->resetvideo, &sleep_time, packet_flags);
					mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "video_decoder_worker_thread audiovisualization_process END st:%d ctx:%8.8X", stream_index, (unsigned int)fvdt->video_ctx);
				}
				if(fvdt->resetoutqueue)
				{
					if((stream_index == MPXPLAY_STREAMTYPEINDEX_VIDEO) || (stream_index == MPXPLAY_STREAMTYPEINDEX_SEEKPREVIEW))
						mpxplay_ffmpgdec_packetqueue_put(&fvdt->videoframe_queue, MPXPLAY_PACKETTYPE_AVFRAME_VIDEO, MPXPLAY_PACKETLISTFLAG_CLEARBUF, NULL, NULL);
					fvdt->resetoutqueue = false;
					mpxplay_debugf(DISPQT_DEBUGOUT_CLEAR, "RESETVIDEO END st:%d ctx:%8.8X q:%d sl:%d", stream_index, (unsigned int)fvdt->video_ctx, fvdt->videoframe_queue.nb_packets, sleep_time);
				}
			}
			else
			{   // currently we simply drop non primary video packets and primary video packet in videowall mode for MPXPLAY_STREAMTYPEINDEX_VIDEO (later can be video crossfade)
				if(fvdt->video_pktlist_elem)
				{
					mpxplay_ffmpgdec_queuelistelem_clear(fvdt->video_pktlist_elem);
					fvdt->video_pktlist_elem = NULL;
				}
				fvdt->resetvideo = fvdt->resetfilter = false;
				mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "video_decoder_worker_thread DROP END");
			}
		}

		if(videwall_is_enabled)
		{
			videoworkdec_videowall_merge_windows(cbelem);
		}

		if(sleep_time > 0)
		{
			pds_threads_sleep(sleep_time);
		}
	}while(!funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_EXITEVENT));

	for(stream_index = MPXPLAY_STREAMTYPEINDEX_VIDEO; stream_index < MPXPLAY_STREAMTYPEINDEX_NUM; stream_index++)
	{
		struct ffmpegvideo_decoder_s *fvdt = &cbelem->video_decoder_threads[stream_index];
		AVCodecContext *video_ctx = fvdt->video_ctx;
		if(video_ctx && avcodec_is_open(video_ctx) && video_ctx->codec)
			avcodec_flush_buffers(video_ctx);
	}
	mpxplay_debugf(DISPQT_DEBUGOUT_VFCC, "video_decoder_worker_thread END");
	if(cbelem && cbelem->infile_callback)
		funcbit_enable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_THEXITDONE);
	pds_threads_sleep(DISPQT_VIDEO_MUTEX_TIMEOUT);  // required for pds_threads_thread_close -> TerminateThread
	return 0;
}

#endif // MPXPLAY_LINK_ORIGINAL_FFMPEG
